﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Problems.DataAnalysis;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.GoalSeeking {
  [Item("BestSolutionAnalyzer", "An analyzer which identifies the best solution from the SingleObjectiveProcessParameterOptimizationProblem")]
  [StorableType("2CAFF3A8-D2FC-4402-A2F8-DD828465870E")]
  public class BestSolutionAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string RowParameterName = "Row";
    private const string ProblemDataParameterName = "ProblemData";
    private const string QualityParameterName = "Quality";
    private const string MaximizationParameterName = "Maximization";

    public bool EnabledByDefault {
      get { return true; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters[QualityParameterName]; }
    }
    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters[MaximizationParameterName]; }
    }
    public IFixedValueParameter<StringValue> BestSolutionResultNameParameter {
      get { return (IFixedValueParameter<StringValue>)Parameters["BestSolution ResultName"]; }
    }
    public ILookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ILookupParameter<IntValue> RowParameter {
      get { return (ILookupParameter<IntValue>)Parameters[RowParameterName]; }
    }
    public ILookupParameter<IRegressionProblemData> ProblemDataParameter {
      get { return (ILookupParameter<IRegressionProblemData>)Parameters[ProblemDataParameterName]; }
    }

    public BoolValue Maximization {
      get { return MaximizationParameter.ActualValue; }
    }

    public ItemArray<DoubleValue> Quality {
      get { return QualityParameter.ActualValue; }
    }

    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }

    public string BestSolutionResultName {
      get { return BestSolutionResultNameParameter.Value.Value; }
      set { BestSolutionResultNameParameter.Value.Value = value; }
    }

    public BestSolutionAnalyzer() {
      Parameters.Add(new FixedValueParameter<StringValue>("BestSolution ResultName", "The name of the result for storing the best solution.", new StringValue("Best Solution")));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", "The quality of the best known solution."));
      Parameters.Add(new LookupParameter<IntValue>(RowParameterName, "The current row"));
      Parameters.Add(new LookupParameter<IRegressionProblemData>(ProblemDataParameterName, "The problem data"));
      Parameters.Add(new LookupParameter<BoolValue>(MaximizationParameterName, "Specifies whether the problem is a minimization or a maximization problem."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>(QualityParameterName, "The qualities of the individuals in the population"));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", "The result collection"));
    }

    protected BestSolutionAnalyzer(BestSolutionAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestSolutionAnalyzer(this, cloner);
    }

    [StorableConstructor]
    protected BestSolutionAnalyzer(StorableConstructorFlag _) : base(_) { }


    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(MaximizationParameterName))
        Parameters.Add(new LookupParameter<BoolValue>(MaximizationParameterName, "Specifies whether the problem is a minimization or a maximization problem."));
      if (!Parameters.ContainsKey(QualityParameterName))
        Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>(QualityParameterName, "The qualities of the individuals in the population"));
      if (!Parameters.ContainsKey("Results"))
        Parameters.Add(new LookupParameter<ResultCollection>("Results", "The result collection"));
    }

    public override IOperation Apply() {
      IEnumerable<IScope> scopes = new[] { ExecutionContext.Scope };
      var zipped = Quality.Select((x, index) => new { Index = index, x.Value });
      var best = Maximization.Value ? zipped.OrderBy(x => x.Value).First() : zipped.OrderByDescending(x => x.Value).First();
      for (int j = 0; j < QualityParameter.Depth; j++)
        scopes = scopes.SelectMany(x => x.SubScopes);
      IScope currentBestScope = scopes.ToList()[best.Index];
      var bestSolution = (RealVector)currentBestScope.Variables["RealVector"].Value;

      var bestSolutionMatrix = new DoubleMatrix(bestSolution.Length, 3);
      var targetNames = bestSolution.ElementNames.ToList();
      bestSolutionMatrix.RowNames = targetNames;
      bestSolutionMatrix.ColumnNames = new[] { "Estimated value", "Target value", "Deviation" };

      var problemData = ProblemDataParameter.ActualValue;
      var row = RowParameter.ActualValue.Value;

      for (int i = 0; i < bestSolution.Length; ++i) {
        var estimatedValue = bestSolution[i];
        var targetValue = problemData.Dataset.GetDoubleValue(targetNames[i], row);
        bestSolutionMatrix[i, 0] = estimatedValue;
        bestSolutionMatrix[i, 1] = targetValue;
        bestSolutionMatrix[i, 2] = estimatedValue - targetValue;
      }

      if (!Results.ContainsKey(BestSolutionResultName)) {
        Results.Add(new Result(BestSolutionResultName, bestSolutionMatrix));
      } else {
        var result = Results[BestSolutionResultName];
        result.Value = bestSolutionMatrix;
      }

      return base.Apply();
    }
  }
}
